/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import android.app.Activity
import android.content.Context
import android.net.Uri
import android.os.Bundle
import android.os.Handler
import android.os.Looper
import android.os.ResultReceiver
import com.kakao.sdk.auth.model.AuthType
import com.kakao.sdk.common.util.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.util.SdkLog
import com.kakao.sdk.common.model.*
import com.kakao.sdk.common.util.IntentResolveClient
import java.net.HttpURLConnection

/**
 * OAuth 2.0 Spec 의 인증 코드를 발급받기 위해 사용되는 클라이언트.
 */
class AuthCodeClient (
    private val intentResolveClient: IntentResolveClient = IntentResolveClient.instance,
    private val applicationInfo: ApplicationInfo = KakaoSdk.applicationContextInfo,
    private val contextInfo: ContextInfo = KakaoSdk.applicationContextInfo
) {

    /**
     * 카카오톡으로 로그인 가능(설치) 여부 검사.
     */
    fun isKakaoTalkLoginAvailable(context: Context): Boolean =
        intentResolveClient.resolveTalkIntent(context, AuthCodeIntentFactory.talkBase()) != null

    /**
     * 카카오톡에 연결된 카카오계정으로 사용자를 인증하고 인증 코드 발급.
     *
     * @param context 카카오톡 로그인 Activity를 실행하기 위한 현재 Activity context
     * @param callback 인증 코드 반환. 발급 받은 인증 코드 는 [AuthApiClient.issueAccessToken] 파라미터로 사용.
     */
    @JvmOverloads
    fun authorizeWithKakaoTalk(
        context: Context,
        requestCode: Int = DEFAULT_REQUEST_CODE,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        callback: (code: String?, error: Throwable?) -> Unit
    ) {
        if (!isKakaoTalkLoginAvailable(context)) {
            callback(null, ClientError(ClientErrorCause.NotSupported, "KakaoTalk not installed"))
        }
        else {
            val clientId = applicationInfo.appKey
            val redirectUri = "kakao$clientId://oauth"
            val kaHeader = contextInfo.kaHeader
            val extras = Bundle().apply {
                channelPublicIds?.let { putString(Constants.CHANNEL_PUBLIC_ID, channelPublicIds.joinToString(",")) }
                serviceTerms?.let { putString(Constants.SERVICE_TERMS, serviceTerms.joinToString(",")) }
            }
            try {
                context.startActivity(
                    AuthCodeIntentFactory.talk(
                        context,
                        requestCode,
                        clientId,
                        redirectUri,
                        kaHeader,
                        extras,
                        resultReceiver(callback)
                    )
                )
            } catch (startActivityError: Throwable) {
                SdkLog.e(startActivityError)
                callback(null, startActivityError)
            }
        }
    }

    /**
     * 기본 웹 브라우저(CustomTabs)에 있는 카카오계정 cookie 로 사용자를 인증하고 인증 코드 발급.
     *
     * @param context CustomTabs를 실행하기 위한 현재 Activity context
     * @param authType 인증 동작 지정. [AuthType]
     * @param scopes 추가로 동의 받고자 하는 동의 항목 ID 목록. 카카오 디벨로퍼스 동의 항목 설정 화면에서 확인 가능.
     * @param callback 인증 코드 반환. 발급 받은 인증 코드는 [AuthApiClient.issueAccessToken] 파라미터로 사용.
     */
    @JvmOverloads
    fun authorizeWithKakaoAccount(
        context: Context,
        authType: AuthType? = null,
        scopes: List<String>? = null,
        agt: String? = null,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        forceAccountLogin: Boolean = false,
        accountParameters: Map<String, String>? = null,
        callback: (code: String?, error: Throwable?) -> Unit
    ) {
        val clientId = applicationInfo.appKey
        val redirectUri = "kakao$clientId://oauth"
        val kaHeader = contextInfo.kaHeader

        val uriUtility = UriUtility()
        val uri =
            uriUtility.authorize(clientId, agt, redirectUri, scopes, kaHeader, channelPublicIds, serviceTerms, authType)
                .let { authorizeUri ->
                    if (forceAccountLogin && accountParameters != null)
                        uriUtility.accountLoginAndAuthorize(authorizeUri, accountParameters)
                    else
                        authorizeUri
                }
        SdkLog.i(uri)
        try {
            context.startActivity(
                AuthCodeIntentFactory.account(
                    context, uri, redirectUri, resultReceiver(callback)
                )
            )
        } catch (startActivityError: Throwable) {
            SdkLog.e(startActivityError)
            callback(null, startActivityError)
        }
    }

    @JvmSynthetic
    internal fun resultReceiver(
        callback: (code: String?, error: Throwable?) -> Unit
    ): ResultReceiver =
        object: ResultReceiver(Handler(Looper.getMainLooper())) {
            override fun onReceiveResult(resultCode: Int, resultData: Bundle?) {
                SdkLog.d("***** AUTH CODE RESULT: $resultData")
                if (resultCode == Activity.RESULT_OK) {
                    val uri = resultData?.getParcelable<Uri>(Constants.KEY_URL)
                    val code = uri?.getQueryParameter(Constants.CODE)
                    if (code != null) {
                        callback(code, null)
                    } else {
                        val error = uri?.getQueryParameter(Constants.ERROR) as String
                        val errorDescription = uri.getQueryParameter(Constants.ERROR_DESCRIPTION) as String
                        callback(null, AuthError(
                            HttpURLConnection.HTTP_MOVED_TEMP,
                            kotlin.runCatching {
                                KakaoJson.fromJson<AuthErrorCause>(error, AuthErrorCause::class.java)
                            }.getOrDefault(
                                AuthErrorCause.Unknown
                            ),
                            AuthErrorResponse(error, errorDescription)
                        ))
                    }

                } else if (resultCode == Activity.RESULT_CANCELED) {
                    var error = resultData?.getSerializable(Constants.KEY_EXCEPTION) as KakaoSdkError
                    callback(null, error)
                } else {
                    var error = IllegalArgumentException("Unknown resultCode in RxAuthCodeClient#onReceivedResult()")
                    callback(null, error)
                }
            }
        }

    companion object {

        /**
         * 간편한 API 호출을 위해 기본 제공되는 singleton 객체
         */
        @JvmStatic
        val instance by lazy { AuthCodeClient() }

        /**
         * 카카오톡 간편로그인 startActivity 에서 사용되는 requestCode 기본 값
         */
        const val DEFAULT_REQUEST_CODE: Int = 10012
    }
}